@@ -43,10 +43,14 @@ except ImportError:
     from StringIO import StringIO
 
 import filters
+import utils
 from generic import *
 from utils import readNonWhitespace, readUntilWhitespace, ConvertFunctionsToVirtualList
 from sets import ImmutableSet
 
+##
+# This class supports writing PDF files out, given pages produced by another
+# class (typically {@link #PdfFileReader PdfFileReader}).
 class PdfFileWriter(object):
     def __init__(self):
         self._header = "%PDF-1.3"
@@ -84,13 +88,15 @@ class PdfFileWriter(object):
         assert ido.pdf == self
         return self._objects[ido.idnum - 1]
 
+    ##
+    # Adds a page to this PDF file.  The page is usually acquired from a
+    # {@link #PdfFileReader PdfFileReader} instance.
+    # <p>
+    # Stability: Added in v1.0, will exist for all v1.x releases.
+    #
+    # @param page The page to add to the document.  This argument should be
+    #             an instance of {@link #PageObject PageObject}.
     def addPage(self, page):
-        """
-        Adds a page to this PDF file.  A dictionary of /Type = /Page.
-        Currently usually aquired from PdfFileReader.getPage().
-
-        Stability: Added in v1.0, will exist for all v1.x releases.
-        """
         assert page["/Type"] == "/Page"
         page[NameObject("/Parent")] = self._pages
         page = self._addObject(page)
@@ -98,13 +104,59 @@ class PdfFileWriter(object):
         pages["/Kids"].append(page)
         pages["/Count"] = NumberObject(pages["/Count"] + 1)
 
-    def write(self, stream):
-        """
-        Writes this PDF file to an output stream.  Writes the file as a
-        PDF-1.3 format file.
+    ##
+    # Encrypt this PDF file with the PDF Standard encryption handler.
+    # @param user_pwd The "user password", which allows for opening and reading
+    # the PDF file with the restrictions provided.
+    # @param owner_pwd The "owner password", which allows for opening the PDF
+    # files without any restrictions.  By default, the owner password is the
+    # same as the user password.
+    # @param use_128bit Boolean argument as to whether to use 128bit
+    # encryption.  When false, 40bit encryption will be used.  By default, this
+    # flag is on.
+    def encrypt(self, user_pwd, owner_pwd = None, use_128bit = True):
+        import md5, time, random
+        if owner_pwd == None:
+            owner_pwd = user_pwd
+        if use_128bit:
+            V = 2
+            rev = 3
+            keylen = 128 / 8
+        else:
+            V = 1
+            rev = 2
+            keylen = 40 / 8
+        # permit everything:
+        P = -1
+        O = StringObject(_alg33(owner_pwd, user_pwd, rev, keylen))
+        ID_1 = md5.new(repr(time.time())).digest()
+        ID_2 = md5.new(repr(random.random())).digest()
+        self._ID = ArrayObject((StringObject(ID_1), StringObject(ID_2)))
+        if rev == 2:
+            U, key = _alg34(user_pwd, O, P, ID_1)
+        else:
+            assert rev == 3
+            U, key = _alg35(user_pwd, rev, keylen, O, P, ID_1, False)
+        encrypt = DictionaryObject()
+        encrypt[NameObject("/Filter")] = NameObject("/Standard")
+        encrypt[NameObject("/V")] = NumberObject(V)
+        if V == 2:
+            encrypt[NameObject("/Length")] = NumberObject(keylen * 8)
+        encrypt[NameObject("/R")] = NumberObject(rev)
+        encrypt[NameObject("/O")] = StringObject(O)
+        encrypt[NameObject("/U")] = StringObject(U)
+        encrypt[NameObject("/P")] = NumberObject(P)
+        self._encrypt = self._addObject(encrypt)
+        self._encrypt_key = key
 
-        Stability: Added in v1.0, will exist for all v1.x releases.
-        """
+    ##
+    # Writes the collection of pages added to this object out as a PDF file.
+    # <p>
+    # Stability: Added in v1.0, will exist for all v1.x releases.
+    # @param stream An object to write the file to.  The object must support
+    # the write method, and the tell method, similar to a file object.
+    def write(self, stream):
+        import struct, md5
 
         externalReferenceMap = {}
         self.stack = []
@@ -115,10 +167,19 @@ class PdfFileWriter(object):
         object_positions = []
         stream.write(self._header + "\n")
         for i in range(len(self._objects)):
+            idnum = (i + 1)
             obj = self._objects[i]
             object_positions.append(stream.tell())
-            stream.write(str(i + 1) + " 0 obj\n")
-            obj.writeToStream(stream)
+            stream.write(str(idnum) + " 0 obj\n")
+            key = None
+            if idnum != self._encrypt.idnum and hasattr(self, "_encrypt_key"):
+                pack1 = struct.pack("<i", i + 1)[:3]
+                pack2 = struct.pack("<i", 0)[:2]
+                key = self._encrypt_key + pack1 + pack2
+                assert len(key) == (len(self._encrypt_key) + 5)
+                md5_hash = md5.new(key).digest()
+                key = md5_hash[:min(16, len(self._encrypt_key) + 5)]
+            obj.writeToStream(stream, key)
             stream.write("\nendobj\n")
 
         # xref table
@@ -137,7 +198,11 @@ class PdfFileWriter(object):
                 NameObject("/Root"): self._root,
                 NameObject("/Info"): self._info,
                 })
-        trailer.writeToStream(stream)
+        if hasattr(self, "_ID"):
+            trailer[NameObject("/ID")] = self._ID
+        if hasattr(self, "_encrypt"):
+            trailer[NameObject("/Encrypt")] = self._encrypt
+        trailer.writeToStream(stream, None)
         
         # eof
         stream.write("\nstartxref\n%s\n%%%%EOF\n" % (xref_location))
@@ -195,30 +260,32 @@ class PdfFileWriter(object):
             return data
 
 
+##
+# Initializes a PdfFileReader object.  This operation can take some time, as
+# the PDF stream's cross-reference tables are read into memory.
+# <p>
+# Stability: Added in v1.0, will exist for all v1.x releases.
+#
+# @param stream An object that supports the standard read and seek methods
+#               similar to a file object.
 class PdfFileReader(object):
     def __init__(self, stream):
-        """
-        Initializes a PdfFileReader object.  This operation can take some time,
-        as the PDF file cross-reference tables are read.  "stream" parameter
-        must be a data stream, not a string or a path name.
-
-        Stability: Added in v1.0, will exist for all v1.x releases.
-        """
         self.flattenedPages = None
         self.resolvedObjects = {}
         self.read(stream)
         self.stream = stream
+        self._override_encryption = False
 
+    ##
+    # Retrieves the PDF file's document information dictionary, if it exists.
+    # Note that some PDF files use metadata streams instead of docinfo
+    # dictionaries, and these metadata streams will not be accessed by this
+    # function.
+    # <p>
+    # Stability: Added in v1.6, will exist for all future v1.x releases.
+    # @return Returns a {@link #DocumentInformation DocumentInformation}
+    #         instance, or None if none exists.
     def getDocumentInfo(self):
-        """
-        Retrieves the PDF file's document information dictionary, if it
-        exists.  Returns a DocumentInformation instance, or None.
-        Note that some PDF files use metadata streams instead of docinfo
-        dictionaries, and these metadata streams will not be accessed by this
-        function.
-
-        Stability: Added in v1.6, will exist for all v1.x releases.
-        """
         if not self.trailer.has_key("/Info"):
             return None
         obj = self.getObject(self.trailer['/Info'])
@@ -226,40 +293,50 @@ class PdfFileReader(object):
         retval.update(obj)
         return retval
 
-    documentInfo = property(lambda self: self.getDocumentInfo(), None, None,
-            """See PdfFileReader.getDocumentInfo().  This property was added 
-            in pyPdf v1.7, and will exist for all future v1.x releases.""")
+    ##
+    # Read-only property that accesses the {@link
+    # #PdfFileReader.getDocumentInfo getDocumentInfo} function.
+    # <p>
+    # Stability: Added in v1.7, will exist for all future v1.x releases.
+    documentInfo = property(lambda self: self.getDocumentInfo(), None, None)
 
+    ##
+    # Calculates the number of pages in this PDF file.
+    # <p>
+    # Stability: Added in v1.0, will exist for all v1.x releases.
+    # @return Returns an integer.
     def getNumPages(self):
-        """
-        Returns the number of pages in this PDF file.
-
-        Stability: Added in v1.0, will exist for all v1.x releases.
-        """
         if self.flattenedPages == None:
             self._flatten()
         return len(self.flattenedPages)
 
-    numPages = property(lambda self: self.getNumPages(), None, None,
-            """See PdfFileReader.getNamePages().  This property was added in
-            v1.7, and will exist for all future v1.x releases.""")
+    ##
+    # Read-only property that accesses the {@link #PdfFileReader.getNumPages
+    # getNumPages} function.
+    # <p>
+    # Stability: Added in v1.7, will exist for all future v1.x releases.
+    numPages = property(lambda self: self.getNumPages(), None, None)
 
+    ##
+    # Retrieves a page by number from this PDF file.
+    # <p>
+    # Stability: Added in v1.0, will exist for all v1.x releases.
+    # @return Returns a {@link #PageObject PageObject} instance.
     def getPage(self, pageNumber):
-        """
-        Retrieves a page by number from this PDF file.  Returns a PageObject
-        instance.
-
-        Stability: Added in v1.0, will exist for all v1.x releases.
-        """
-        # ensure that we're not trying to access an encrypted PDF
-        assert not self.trailer.has_key("/Encrypt")
+        ## ensure that we're not trying to access an encrypted PDF
+        #assert not self.trailer.has_key("/Encrypt")
         if self.flattenedPages == None:
             self._flatten()
         return self.flattenedPages[pageNumber]
 
+    ##
+    # Read-only property that emulates a list based upon the {@link
+    # #PdfFileReader.getNumPages getNumPages} and {@link #PdfFileReader.getPage
+    # getPage} functions.
+    # <p>
+    # Stability: Added in v1.7, and will exist for all future v1.x releases.
     pages = property(lambda self: ConvertFunctionsToVirtualList(self.getNumPages, self.getPage),
-            None, None, """Returns a sequence of pages.  This property was
-            added in v1.7 and will exist for all future v1.x releases.""")
+            None, None)
 
     def _flatten(self, pages = None, inherit = None):
         inheritablePageAttributes = (
@@ -291,6 +368,11 @@ class PdfFileReader(object):
             pageObj.update(pages)
             self.flattenedPages.append(pageObj)
 
+    def safeGetObject(self, obj):
+        if isinstance(obj, IndirectObject):
+            return self.safeGetObject(self.getObject(obj))
+        return obj
+
     def getObject(self, indirectReference):
         retval = self.resolvedObjects.get(indirectReference.generation, {}).get(indirectReference.idnum, None)
         if retval != None:
@@ -323,9 +405,38 @@ class PdfFileReader(object):
         assert idnum == indirectReference.idnum
         assert generation == indirectReference.generation
         retval = readObject(self.stream, self)
+
+        # override encryption is used for the /Encrypt dictionary
+        if not self._override_encryption:
+            # if we don't have the encryption key:
+            if self.isEncrypted and not hasattr(self, '_decryption_key'):
+                raise Exception, "file has not been decrypted"
+            # otherwise, decrypt here...
+            import struct, md5
+            pack1 = struct.pack("<i", indirectReference.idnum)[:3]
+            pack2 = struct.pack("<i", indirectReference.generation)[:2]
+            key = self._decryption_key + pack1 + pack2
+            assert len(key) == (len(self._decryption_key) + 5)
+            md5_hash = md5.new(key).digest()
+            key = md5_hash[:min(16, len(self._decryption_key) + 5)]
+            retval = self._decryptObject(retval, key)
+
         self.cacheIndirectObject(generation, idnum, retval)
         return retval
 
+    def _decryptObject(self, obj, key):
+        if isinstance(obj, StringObject):
+            obj = StringObject(utils.RC4_encrypt(key, obj))
+        elif isinstance(obj, StreamObject):
+            obj._data = utils.RC4_encrypt(key, obj._data)
+        elif isinstance(obj, DictionaryObject):
+            for dictkey, value in obj.items():
+                obj[dictkey] = self._decryptObject(value, key)
+        elif isinstance(obj, ArrayObject):
+            for i in range(len(obj)):
+                obj[i] = self._decryptObject(obj[i], key)
+        return obj
+
     def readObjectHeader(self, stream):
         idnum = readUntilWhitespace(stream)
         generation = readUntilWhitespace(stream)
@@ -488,6 +599,91 @@ class PdfFileReader(object):
                 line = x + line
         return line
 
+    ##
+    # When using an encrypted / secured PDF file with the PDF Standard
+    # encryption handler, this function will allow the file to be decrypted.
+    # It checks the given password against the document's user password and
+    # owner password, and then stores the resulting decryption key if either
+    # password is correct.
+    # <p>
+    # It does not matter which password was matched.  Both passwords provide
+    # the correct decryption key that will allow the document to be used with
+    # this library.
+    # <p>
+    # Stability: Added in v1.8, will exist for all future v1.x releases.
+    #
+    # @return 0 if the password failed, 1 if the password matched the user
+    # password, and 2 if the password matched the owner password.
+    #
+    # @exception NotImplementedError Document uses an unsupported encryption
+    # method.
+    def decrypt(self, password):
+        self._override_encryption = True
+        try:
+            return self._decrypt(password)
+        finally:
+            self._override_encryption = False
+
+    def _decrypt(self, password):
+        encrypt = self.safeGetObject(self.trailer['/Encrypt'])
+        if encrypt['/Filter'] != '/Standard':
+            raise NotImplementedError, "only Standard PDF encryption handler is available"
+        if not (encrypt['/V'] in (1, 2)):
+            raise NotImplementedError, "only algorithm code 1 and 2 are supported"
+        user_password, key = self._authenticateUserPassword(password)
+        if user_password:
+            self._decryption_key = key
+            return 1
+        else:
+            rev = self.safeGetObject(encrypt['/R'])
+            if rev == 2:
+                keylen = 5
+            else:
+                keylen = self.safeGetObject(encrypt['/Length']) / 8
+            key = _alg33_1(password, rev, keylen)
+            real_O = self.safeGetObject(encrypt["/O"])
+            if rev == 2:
+                userpass = utils.RC4_encrypt(key, real_O)
+            else:
+                val = real_O
+                for i in range(19, -1, -1):
+                    new_key = ''
+                    for l in range(len(key)):
+                        new_key += chr(ord(key[l]) ^ i)
+                    val = utils.RC4_encrypt(new_key, val)
+                userpass = val
+            owner_password, key = self._authenticateUserPassword(userpass)
+            if owner_password:
+                self._decryption_key = key
+                return 2
+        return 0
+
+    def _authenticateUserPassword(self, password):
+        encrypt = self.safeGetObject(self.trailer['/Encrypt'])
+        rev = self.safeGetObject(encrypt['/R'])
+        owner_entry = self.safeGetObject(encrypt['/O'])
+        p_entry = self.safeGetObject(encrypt['/P'])
+        id_entry = self.safeGetObject(self.trailer['/ID'])
+        id1_entry = self.safeGetObject(id_entry[0])
+        if rev == 2:
+            U, key = _alg34(password, owner_entry, p_entry, id1_entry)
+        elif rev >= 3:
+            U, key = _alg35(password, rev,
+                    self.safeGetObject(encrypt["/Length"]) / 8, owner_entry,
+                    p_entry, id1_entry,
+                    self.safeGetObject(encrypt.get("/EncryptMetadata", False)))
+        real_U = self.safeGetObject(encrypt['/U'])
+        return U == real_U, key
+
+    def getIsEncrypted(self):
+        return self.trailer.has_key("/Encrypt")
+
+    ##
+    # Read-only boolean property showing whether this PDF file is encrypted.
+    # Note that this property, if true, will remain true even after the {@link
+    # #PdfFileReader.decrypt decrypt} function is called.
+    isEncrypted = property(lambda self: self.getIsEncrypted(), None, None)
+
 
 def getRectangle(self, name, defaults):
     retval = self.get(name)
@@ -512,38 +708,39 @@ def setRectangle(self, name, value):
 def deleteRectangle(self, name):
     del self[name]
 
-def addRectangleAccessor(klass, propname, name, fallback, docs):
-    setattr(klass, propname,
+def createRectangleAccessor(name, fallback):
+    return \
         property(
             lambda self: getRectangle(self, name, fallback),
             lambda self, value: setRectangle(self, name, value),
-            lambda self: deleteRectangle(self, name),
-            docs
+            lambda self: deleteRectangle(self, name)
             )
-        )
 
+##
+# This class represents a single page within a PDF file.  Typically this object
+# will be created by accessing the {@link #PdfFileReader.getPage getPage}
+# function of the {@link #PdfFileReader PdfFileReader} class.
 class PageObject(DictionaryObject):
     def __init__(self, pdf):
         DictionaryObject.__init__(self)
         self.pdf = pdf
 
+    ##
+    # Rotates a page clockwise by increments of 90 degrees.
+    # <p>
+    # Stability: Added in v1.1, will exist for all future v1.x releases.
+    # @param angle Angle to rotate the page.  Must be an increment of 90 deg.
     def rotateClockwise(self, angle):
-        """
-        Rotates a page clockwise by increments of 90 degrees.
-
-        Stability: Added in v1.1, will exist for all v1.x releases thereafter.
-        """
         assert angle % 90 == 0
         self._rotate(angle)
         return self
 
+    ##
+    # Rotates a page counter-clockwise by increments of 90 degrees.
+    # <p>
+    # Stability: Added in v1.1, will exist for all future v1.x releases.
+    # @param angle Angle to rotate the page.  Must be an increment of 90 deg.
     def rotateCounterClockwise(self, angle):
-        """
-        Rotates a page counter-clockwise by increments of 90 degrees.  Note
-        that this is equivilant to calling rotateClockwise(-angle).
-
-        Stability: Added in v1.1, will exist for all v1.x releases thereafter.
-        """
         assert angle % 90 == 0
         self._rotate(-angle)
         return self
@@ -589,14 +786,17 @@ class PageObject(DictionaryObject):
         return stream
     _pushPopGS = staticmethod(_pushPopGS)
 
+    ##
+    # Merges the content streams of two pages into one.  Resource references
+    # (i.e. fonts) are maintained from both pages.  The mediabox/cropbox/etc
+    # of this page are not altered.  The parameter page's content stream will
+    # be added to the end of this page's content stream, meaning that it will
+    # be drawn after, or "on top" of this page.
+    # <p>
+    # Stability: Added in v1.4, will exist for all future 1.x releases.
+    # @param page2 An instance of {@link #PageObject PageObject} to be merged
+    #              into this one.
     def mergePage(self, page2):
-        """
-        Merges the content streams of two pages into one.  Resource
-        references (i.e. fonts) are maintained from both pages.  The
-        mediabox/cropbox/etc on "self" are not altered.
-
-        Stability: Added in v1.4, will exist for all v1.x releases thereafter.
-        """
 
         # First we work on merging the resource dictionaries.  This allows us
         # to find out what symbols in the content streams we might need to
@@ -633,32 +833,31 @@ class PageObject(DictionaryObject):
         self[NameObject('/Contents')] = ContentStream(newContentArray, self.pdf)
         self[NameObject('/Resources')] = newResources
 
+    ##
+    # Compresses the size of this page by joining all content streams and
+    # applying a FlateDecode filter.
+    # <p>
+    # Stability: Added in v1.6, will exist for all future v1.x releases.
+    # However, it is possible that this function will perform no action if
+    # content stream compression becomes "automatic" for some reason.
     def compressContentStreams(self):
-        """
-        Join all content streams and apply a FlateDecode filter to decrease
-        the stream's size.
-
-        Stability: Added in v1.6, will exist for all v1.x releases thereafter.
-        However, if content stream compression is ever handled in a different
-        and/or more transparent way, this function may not do anything.
-        """
         content = self["/Contents"].getObject()
         if not isinstance(content, ContentStream):
             content = ContentStream(content, self.pdf)
         self[NameObject("/Contents")] = content.flateEncode()
 
+    ##
+    # Locate all text drawing commands, in the order they are provided in the
+    # content stream, and extract the text.  This works well for some PDF
+    # files, but poorly for others, depending on the generator used.  This will
+    # be refined in the future.  Do not rely on the order of text coming out of
+    # this function, as it will change if this function is made more
+    # sophisticated.
+    # <p>
+    # Stability: Added in v1.7, will exist for all future v1.x releases.  May
+    # be overhauled to provide more ordered text in the future.
+    # @return a string object
     def extractText(self):
-        """
-        Locate all text drawing commands, in the order they are provided in
-        the content stream, and extract the text.  This works well for some
-        PDF files, but poorly for others, depending on the generator used.
-        This will be refined in the future.  Do not rely on the order of text
-        coming out of this function, as it will change if this function is 
-        made more sophisticated.
-
-        Stability: Added in v1.7, will exist for all v1.x releases thereafter.
-        May be overhauled to provide more ordered text in the future.
-        """
         text = ""
         content = self["/Contents"].getObject()
         if not isinstance(content, ContentStream):
@@ -680,42 +879,46 @@ class PageObject(DictionaryObject):
                         text += i
         return text
 
-addRectangleAccessor(PageObject, "mediaBox", "/MediaBox", (),
-        """A rectangle, expressed in default user space units, defining the
-        boundaries of the physical medium on which the page is intended to be
-        displayed or printed.
+    ##
+    # A rectangle (RectangleObject), expressed in default user space units,
+    # defining the boundaries of the physical medium on which the page is
+    # intended to be displayed or printed.
+    # <p>
+    # Stability: Added in v1.4, will exist for all future v1.x releases.
+    mediaBox = createRectangleAccessor("/MediaBox", ())
 
-        Stability: Added in v1.4, will exist for all v1.x releases
-        thereafter.""")
-addRectangleAccessor(PageObject, "cropBox", "/CropBox", ("/MediaBox",),
-        """A rectangle, expressed in default user space units, defining the
-        visible region of default user space.  When the page is displayed or
-        printed, its contents are to be clipped (cropped) to this rectangle and
-        then imposed on the output medium in some implementation-defined
-        manner.  Default value: same as MediaBox.
+    ##
+    # A rectangle (RectangleObject), expressed in default user space units,
+    # defining the visible region of default user space.  When the page is
+    # displayed or printed, its contents are to be clipped (cropped) to this
+    # rectangle and then imposed on the output medium in some
+    # implementation-defined manner.  Default value: same as MediaBox.
+    # <p>
+    # Stability: Added in v1.4, will exist for all future v1.x releases.
+    cropBox = createRectangleAccessor("/CropBox", ("/CropBox",))
 
-        Stability: Added in v1.4, will exist for all v1.x releases
-        thereafter.""")
-addRectangleAccessor(PageObject, "bleedBox", "/BleedBox", ("/CropBox",
-        "/MediaBox"), """A rectangle, expressed in default user space units,
-        defining the region to which the contents of the page should be clipped
-        when output in a production environment.
-        
-        Stability: Added in v1.4, will exist for all v1.x releases
-        thereafter.""")
-addRectangleAccessor(PageObject, "trimBox", "/TrimBox", ("/CropBox",
-        "/MediaBox"), """A rectangle, expressed in default user space units,
-        defining the intended dimensions of the finished page after trimming.
-        
-        Stability: Added in v1.4, will exist for all v1.x releases
-        thereafter.""")
-addRectangleAccessor(PageObject, "artBox", "/ArtBox", ("/CropBox",
-        "/MediaBox"), """A rectangle, expressed in default user space units,
-        defining the extent of the page's meaningful content as intended by the
-        page's creator.
-        
-        Stability: Added in v1.4, will exist for all v1.x releases
-        thereafter.""")
+    ##
+    # A rectangle (RectangleObject), expressed in default user space units,
+    # defining the region to which the contents of the page should be clipped
+    # when output in a production enviroment.
+    # <p>
+    # Stability: Added in v1.4, will exist for all future v1.x releases.
+    bleedBox = createRectangleAccessor("/BleedBox", ("/CropBox", "/MediaBox"))
+
+    ##
+    # A rectangle (RectangleObject), expressed in default user space units,
+    # defining the intended dimensions of the finished page after trimming.
+    # <p>
+    # Stability: Added in v1.4, will exist for all future v1.x releases.
+    trimBox = createRectangleAccessor("/TrimBox", ("/CropBox", "/MediaBox"))
+
+    ##
+    # A rectangle (RectangleObject), expressed in default user space units,
+    # defining the extent of the page's meaningful content as intended by the
+    # page's creator.
+    # <p>
+    # Stability: Added in v1.4, will exist for all future v1.x releases.
+    artBox = createRectangleAccessor("/ArtBox", ("/CropBox", "/MediaBox"))
 
 
 class ContentStream(DecodedStreamObject):
@@ -816,42 +1019,45 @@ class ContentStream(DecodedStreamObject):
     _data = property(_getData, _setData)
 
 
+##
+# A class representing the basic document metadata provided in a PDF File.
 class DocumentInformation(DictionaryObject):
     def __init__(self):
         DictionaryObject.__init__(self)
 
-    title = property(
-            lambda self: self.get("/Title", None),
-            None, None,
-            """The document's title, or None if not specified.  Added to pyPdf
-            in v1.6, will exist for all v1.x.""")
+    ##
+    # Read-only property accessing the document's title.  Added in v1.6, will
+    # exist for all future v1.x releases.
+    # @return A string, or None if the title is not provided.
+    title = property(lambda self: self.get("/Title", None), None, None)
 
-    author = property(
-            lambda self: self.get("/Author", None),
-            None, None,
-            """The name of the person who created the document, or None if not
-            specified.  Added to pyPdf in v1.6, will exist for all v1.x.""")
+    ##
+    # Read-only property accessing the document's author.  Added in v1.6, will
+    # exist for all future v1.x releases.
+    # @return A string, or None if the author is not provided.
+    author = property(lambda self: self.get("/Author", None), None, None)
 
-    subject = property(
-            lambda self: self.get("/Subject", None),
-            None, None,
-            """The subject of the document, or None if not specified.  Added to
-            pyPdf in v1.6, will exist for all v1.x.""")
+    ##
+    # Read-only property accessing the subject of the document.  Added in v1.6,
+    # will exist for all future v1.x releases.
+    # @return A string, or None if the subject is not provided.
+    subject = property(lambda self: self.get("/Subject", None), None, None)
 
-    creator = property(
-            lambda self: self.get("/Creator", None),
-            None, None,
-            """If the document was converted to PDF from another format, the
-            name of the application (for example, OpenOffice) that created the
-            original document from which it was converted, or None if not
-            specified.  Added to pyPdf in v1.6, will exist for all v1.x.""")
+    ##
+    # Read-only property accessing the document's creator.  If the document was
+    # converted to PDF from another format, the name of the application (for
+    # example, OpenOffice) that created the original document from which it was
+    # converted.  Added in v1.6, will exist for all future v1.x releases.
+    # @return A string, or None if the creator is not provided.
+    creator = property(lambda self: self.get("/Creator", None), None, None)
 
-    producer = property(
-            lambda self: self.get("/Producer", None),
-            None, None,
-            """If the document was converted to PDF from another format, the
-            name of the application (for example, OSX Quartz) that converted it
-            to PDF.  Added to pyPdf in v1.6, will exist for all v1.x.""")
+    ##
+    # Read-only property accessing the document's producer.  If the document
+    # was converted to PDF from another format, the name of the application
+    # (for example, OSX Quartz) that converted it to PDF.  Added in v1.6, will
+    # exist for all future v1.x releases.
+    # @return A string, or None if the producer is not provided.
+    producer = property(lambda self: self.get("/Producer", None), None, None)
 
 
 def convertToInt(d, size):
@@ -867,25 +1073,90 @@ def convertToInt(d, size):
         # size too big
         assert False
 
+# ref: pdf1.8 spec section 3.5.2 algorithm 3.2
+_encryption_padding = '\x28\xbf\x4e\x5e\x4e\x75\x8a\x41\x64\x00\x4e\x56' + \
+        '\xff\xfa\x01\x08\x2e\x2e\x00\xb6\xd0\x68\x3e\x80\x2f\x0c' + \
+        '\xa9\xfe\x64\x53\x69\x7a'
 
-if __name__ == "__main__":
-    output = PdfFileWriter()
+def _alg32(password, rev, keylen, owner_entry, p_entry, id1_entry, metadata_encrypt=True):
+    import md5, struct
+    m = md5.new()
+    password = (password + _encryption_padding)[:32]
+    m.update(password)
+    m.update(owner_entry)
+    p_entry = struct.pack('<i', p_entry)
+    m.update(p_entry)
+    m.update(id1_entry)
+    if rev >= 3 and not metadata_encrypt:
+        m.update("\xff\xff\xff\xff")
+    md5_hash = m.digest()
+    if rev >= 3:
+        for i in range(50):
+            md5_hash = md5.new(md5_hash[:keylen]).digest()
+    return md5_hash[:keylen]
 
-    input1 = PdfFileReader(file("test\\5000-s1-05e.pdf", "rb"))
-    page1 = input1.getPage(0)
+def _alg33(owner_pwd, user_pwd, rev, keylen):
+    key = _alg33_1(owner_pwd, rev, keylen)
+    user_pwd = (user_pwd + _encryption_padding)[:32]
+    val = utils.RC4_encrypt(key, user_pwd)
+    if rev >= 3:
+        for i in range(1, 20):
+            new_key = ''
+            for l in range(len(key)):
+                new_key += chr(ord(key[l]) ^ i)
+            val = utils.RC4_encrypt(new_key, val)
+    return val
 
-    input2 = PdfFileReader(file("test\\PDFReference16.pdf", "rb"))
-    page2 = input2.getPage(0)
-    page3 = input2.getPage(1)
-    page1.mergePage(page2)
-    page1.mergePage(page3)
+def _alg33_1(password, rev, keylen):
+    import md5
+    m = md5.new()
+    password = (password + _encryption_padding)[:32]
+    m.update(password)
+    md5_hash = m.digest()
+    if rev >= 3:
+        for i in range(50):
+            md5_hash = md5.new(md5_hash).digest()
+    key = md5_hash[:keylen]
+    return key
 
-    input3 = PdfFileReader(file("test\\cc-cc.pdf", "rb"))
-    page1.mergePage(input3.getPage(0))
+def _alg34(password, owner_entry, p_entry, id1_entry):
+    key = _alg32(password, 2, 5, owner_entry, p_entry, id1_entry)
+    U = utils.RC4_encrypt(key, _encryption_padding)
+    return U, key
 
-    page1.compressContentStreams()
+def _alg35(password, rev, keylen, owner_entry, p_entry, id1_entry, metadata_encrypt):
+    import md5
+    m = md5.new()
+    m.update(_encryption_padding)
+    m.update(id1_entry)
+    md5_hash = m.digest()
+    key = _alg32(password, rev, keylen, owner_entry, p_entry, id1_entry)
+    val = utils.RC4_encrypt(key, md5_hash)
+    for i in range(1, 20):
+        new_key = ''
+        for l in range(len(key)):
+            new_key += chr(ord(key[l]) ^ i)
+        val = utils.RC4_encrypt(new_key, val)
+    return val + ('\x00' * 16), key
 
-    output.addPage(page1)
-    output.write(file("test\\merge-test.pdf", "wb"))
+#if __name__ == "__main__":
+#    output = PdfFileWriter()
+#
+#    input1 = PdfFileReader(file("test\\5000-s1-05e.pdf", "rb"))
+#    page1 = input1.getPage(0)
+#
+#    input2 = PdfFileReader(file("test\\PDFReference16.pdf", "rb"))
+#    page2 = input2.getPage(0)
+#    page3 = input2.getPage(1)
+#    page1.mergePage(page2)
+#    page1.mergePage(page3)
+#
+#    input3 = PdfFileReader(file("test\\cc-cc.pdf", "rb"))
+#    page1.mergePage(input3.getPage(0))
+#
+#    page1.compressContentStreams()
+#
+#    output.addPage(page1)
+#    output.write(file("test\\merge-test.pdf", "wb"))
 
 
